#!/usr/bin/env python3
"""Log events."""

import logging
import os


def _logger_config():
    """Configure log format."""
    if os.environ.get("DEBUG"):
        log_format = "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s"
        level = logging.DEBUG
    else:
        log_format = "%(asctime)s [%(levelname)s] %(message)s"
        level = logging.INFO
    return {"level": level, "format": log_format, "filemode": "w"}


def logger(name="tolkein"):
    """
    Create logger.

    Args:
        name (str, optional): Logger name. Defaults to "tolkein".

    Returns:
        logging.Logger: A logger instance.
    """
    config = _logger_config()
    logging.basicConfig(**config)
    _logger = logging.getLogger(name)
    _logger.propagate = False
    # _logger.handlers = []
    _logger.setLevel(config["level"])
    stream_h = logging.StreamHandler()
    formatter = logging.Formatter(config["format"])
    stream_h.setFormatter(formatter)
    _logger.addHandler(stream_h)
    for handler in _logger.handlers:
        handler.formatter.default_msec_format = "%s.%03d"
    return _logger


class DisableLogger:
    """
    Logger context management.

    .. testsetup::

       from tolkein.lib.tolog import logger
       from tolkein.lib.tolog import DisableLogger
       my_logger = logger(__name__)

    .. doctest::

        >>> my_logger.info('Print log messages')
        >>> with DisableLogger():
        ...     my_logger.info('Disable log messages')
        >>> my_logger.info('Print log messages again')

    """

    def __enter__(self):
        """Set logging level to critical."""
        logging.disable(logging.CRITICAL)

    def __exit__(self, x, y, z):
        """Set logging level back to default."""
        logging.disable(logging.NOTSET)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python3
"""
Tolkein.

usage: tolkein [<command>] [<args>...] [-h|--help] [--version]

commands:
    -h, --help      show this
    -v, --version   show version number
"""

import sys
from importlib.metadata import entry_points

from docopt import DocoptExit
from docopt import docopt

from .lib.tolog import logger
from .lib.version import __version__

LOGGER = logger(__name__)


def cli():
    """Entry point."""
    if len(sys.argv) > 1:
        try:
            args = docopt(__doc__, help=False, version=__version__)
        except DocoptExit:
            args = {"<command>": sys.argv[1]}
        if args["<command>"]:
            # load <command> from entry_points
            eps = entry_points()
            if hasattr(eps, "select"):
                # Python 3.10+
                group = eps.select(group="tolkein.subcmd")
            else:
                # Python 3.9
                group = eps.get("tolkein.subcmd", [])
            for entry_point in group:
                if entry_point.name == args["<command>"]:
                    subcommand = entry_point.load()
                    sys.exit(subcommand(sys.argv[1:]))
            LOGGER.error("'tolkein %s' is not a valid command", args["<command>"])
            sys.exit(1)
    print(__doc__)
    raise DocoptExit



#!/usr/bin/env python3
"""Wikidata functions."""

import re
from collections import defaultdict
from datetime import datetime

from SPARQLWrapper import JSON
from SPARQLWrapper import SPARQLWrapper
from tqdm import tqdm

from genomehubs.vendor.tolkein import tofile
from genomehubs.vendor.tolkein import tolog

LOGGER = tolog.logger(__name__)

SPARQL = SPARQLWrapper(
    "https://query.wikidata.org/sparql",
    agent="Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.11 (KHTML, like Gecko) Chrome/23.0.1271.64 Safari/537.11",
)
WD = "http://www.wikidata.org/entity/"
ITEM = re.compile(r"^Q\d+$")
PROP = re.compile(r"(<|>|\"|direct\/)")
LAST = datetime.now()

RANKS = [
    "subspecies",
    "species",
    "genus",
    "family",
    "order",
    "class",
    "subphylum",
    "phylum",
    "kingdom",
    "superkingdom",
]

# PROPERTIES = {
#     "P225": taxon name
#     "P105": taxon rank
#     "P171": parent taxon
#     "P3606": BOLD Systems taxon ID
#     "P846": GBIF taxonKey
#     "P3240": NBN System Key
#     "P685": NCBI taxonomy ID
# }

SOURCES = {
    "bold": {
        "property": "P3606",
        "source": "BOLD",
        "display_name": "BOLD Systems taxon ID",
        "stub": "http://www.boldsystems.org/index.php/TaxBrowser_TaxonPage?taxid=",
    },
    "gbif": {
        "property": "P846",
        "source": "GBIF",
        "display_name": "GBIF taxonKey",
        "stub": "https://www.gbif.org/species/",
    },
    "nbn": {
        "property": "P3240",
        "source": "NBN",
        "display_name": "NBN System Key",
        "stub": "https://data.nbn.org.uk/Taxa/",
    },
    "ncbi": {
        "property": "P685",
        "source": "NCBI",
        "display_name": "NCBI taxonomy ID",
        "stub": "https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?mode=Info&id=",
    },
    "wikidata": {
        "property": None,
        "source": "Wikidata",
        "display_name": "Wikidata entity",
        "stub": "https://www.wikidata.org/wiki/",
    },
}

# useful wikidata entities:
# Q16521 = taxon
# Q427626 = taxonomic rank

# useful wikidata properties:
# P31 = instance of
# P225 = taxon name
# P105 = taxon rank
# P171 = parent taxon
# P3606 = BOLD Systems taxon ID
# P846 = GBIF taxonKey
# P3240 = NBN System Key
# P685 = NCBI taxonomy ID

# grep -w 'P225\|P105\|P171\|P3606\|P846\|P3240\|P685'


def load_wikidata_dump(filename, rank_entities, *, roots=None):
    """Load data from a wikidata dump file."""
    props = defaultdict(dict)
    return_props = {}
    with tofile.open_file_handle(filename) as fh:
        LOGGER.info("Reading wikidata triples")
        for line in tqdm(fh):
            parts = PROP.split(line)
            if len(parts) != 15:
                continue
            entity = parts[2]
            prop = parts[8]
            value = parts[12]
            props[entity].update({prop: value})
    LOGGER.info("Inferring lineages")
    for entity, entity_props in tqdm(props.items()):
        # if "lineage" in entity_props and entity_props["lineage"]:
        #     if "target" in entity_props["lineage"]:
        #         includes_target = True
        #     if roots is None or includes_target:
        #         return_props[entity] = props[entity]
        #     continue
        lineage = {}
        ancestors = set()
        # entity_props["lineage"] = {}
        # lineages = [entity_props["lineage"]]
        includes_target = False
        parent = entity_props.get("P171", None)
        while parent is not None:
            if parent not in props or parent == entity or parent in ancestors:
                break
            ancestors.add(parent)
            parent_props = props[parent]
            parent_rank = parent_props.get("P105", None)
            if (
                parent_rank is not None
                and parent_rank in rank_entities
                and rank_entities[parent_rank] in RANKS
            ):
                parent_name = parent_props.get("P225", None)
                if parent_name is not None:
                    # for lineage in lineages:
                    lineage.update({rank_entities[parent_rank]: parent_name})
                    if roots is not None and parent_name in roots:
                        includes_target = True
                        lineage.update({"target": True})
            if "lineage" in parent_props and parent_props["lineage"]:
                if "target" in parent_props["lineage"]:
                    includes_target = True
                # for lineage in lineages:
                lineage = {**lineage, **parent_props["lineage"]}
                break
            # else:
            #     parent_props["lineage"] = {}
            #     lineages.append(parent_props["lineage"])
            parent = parent_props.get("P171", None)
        if lineage:
            entity_props["lineage"] = lineage
        # if not entity_props["lineage"]:
        entity_name = entity_props.get("P225", "")
        if entity_name in roots:
            includes_target = True
        if roots is None or includes_target:
            return_props[entity] = props[entity]
    return return_props


def fetch_wikidata_rank_entities():
    """Fetch wikidata entities for given rank names."""
    SPARQL.setQuery(
        """
SELECT ?item ?itemLabel WHERE {
  ?item wdt:P31 wd:Q427626 .
  SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
}
"""
    )
    SPARQL.setReturnFormat(JSON)
    data = SPARQL.query().convert()
    results = data["results"]["bindings"]
    entities = {}
    ranks = {}
    for result in results:
        entities.update({result["item"]["value"]: result["itemLabel"]["value"]})
        ranks.update({result["itemLabel"]["value"]: result["item"]["value"]})
    return entities, ranks


def prepare_xref_rows(key, meta, entities):
    """Convert identifiers to a set of rows for output."""
    ranks = [
        "subspecies",
        "species",
        "genus",
        "family",
        "order",
        "class",
        "subphylum",
        "phylum",
    ]
    try:
        lineage = meta["lineage"]
    except KeyError:
        lineage = {}

    rows = []
    common = {}
    for db in SOURCES.keys():
        common.update({db: "None"})
    entity = key.replace(WD, "")
    for rank in ranks:
        if rank in lineage:
            common.update({rank: lineage[rank]})
    if SOURCES["ncbi"]["property"] in meta:
        common.update({"ncbiTaxonId": meta[SOURCES["ncbi"]["property"]]})
        common.update({"taxonId": meta[SOURCES["ncbi"]["property"]]})
    else:
        common.update({"taxonId": entity})
    if "P225" in meta:
        name = meta["P225"]
        if "P105" in meta and meta["P105"] in entities:
            rank = entities[meta["P105"]]
            common.update({rank: name})
    common.update({"wikidataTaxonId": entity})
    row = {**common}
    for db in SOURCES.keys():
        if SOURCES[db]["property"] is None or SOURCES[db]["property"] in meta:
            row = {**common}
            if SOURCES[db]["property"] is None:
                slug = entity
            else:
                slug = str(meta[SOURCES[db]["property"]])
            row.update(
                {
                    db: slug,
                    "source": SOURCES[db]["source"],
                    "sourceStub": SOURCES[db]["stub"],
                }
            )
            rows.append(row)
    return rows


def wikidata_parser(_params, opts, *, types=None, names=None):
    """Parse WikiData taxa and identifiers."""
    if names is None:
        names = {}
    if "taxon_names" not in names:
        names["taxon_names"] = {}
    for db, values in SOURCES.items():
        names["taxon_names"][db] = {
            "display_name": values["display_name"],
            "header": db,
            "xref": True,
        }
    parsed = []
    entities, ranks = fetch_wikidata_rank_entities()
    roots = opts.get("wikidata-root", None)
    data = load_wikidata_dump(opts["wikidata"], entities, roots=roots)
    for key, meta in data.items():
        parsed += prepare_xref_rows(key, meta, entities)
    return parsed
    return parsed



#!/usr/bin/env python3

# pylint: disable=c-extension-no-member

"""Read, write and parse files."""

import csv
import gzip
import io
import os
import pathlib
import sys
from itertools import groupby
from subprocess import PIPE
from subprocess import Popen

import ujson
import yaml


def open_file_handle(filename):
    """
    Open a filehandle.

    Automatically detect gzipped files based on suffix.

    Args:
        filename (str): Name of file to read.

    Returns:
        An open filehandle. Will return None if the file cannot be opened.
    """
    if ".gz" in pathlib.Path(filename).suffixes:
        try:
            return gzip.open(filename, "rt")
        except OSError:
            return None
    try:
        return open(filename, "r")
    except IOError:
        return None


def read_file(filename):
    """
    Read a whole file into memory.

    Automatically detect gzipped files based on suffix.

    Args:
        filename (str): Name of file to read.

    Returns:
        str: Content of file as a string. Will return None if file cannot be read.
    """
    try:
        with open_file_handle(filename) as fh:
            return fh.read()
    except AttributeError:
        return None


def stream_fasta(filename):
    """
    Stream a FASTA file, sequence by sequence.

    Automatically detect gzipped files based on suffix.

    Args:
        filename (str): Name of FASTA file to read.

    Yields:
        A tuple of::

            (
                str: Sequence ID,
                str: Sequence string
            )
    """
    if ".gz" in pathlib.Path(filename).suffixes:
        cmd = ["pigz", "-dc", filename]
    else:
        cmd = ["cat", filename]
    with Popen(cmd, stdout=PIPE, encoding="utf-8", bufsize=4096) as proc:
        faiter = (x[1] for x in groupby(proc.stdout, lambda line: line[0] == ">"))
        for header in faiter:
            seq_id = header.__next__().split()[0].replace(">", "")
            seq_str = "".join(map(lambda s: s.strip(), faiter.__next__()))
            yield seq_id, seq_str


def load_yaml(filename):
    """
    Parse a JSON/YAML file.

    Args:
        filename (str): Name of JSON/YAML file to parse.

    Returns:
        Dict or list of file content.
    """
    data = read_file(filename)
    if data is None:
        return data
    if ".json" in filename:
        content = ujson.loads(data)
    elif ".yaml" in filename or ".yml" in filename:
        content = yaml.full_load(data)
    else:
        content = data
    return content


def write_file(filename, data, *, plain=False):
    """
    Write a file, use suffix to determine type and compression.

    - types: '.json', '.yaml'
    - compression: None, '.gz'

    Args:
        filename (str): Name of FASTA file to read.
        data: data to write to file.
        plain (bool, optional): Whether to treat data as plain text. Defaults to False.

    Returns:
        bool: Whether file was written successfully.
    """
    if ".json" in filename:
        content = ujson.dumps(data, indent=1, escape_forward_slashes=False)
    elif ".yaml" in filename or ".yml" in filename:
        content = yaml.dump(data, indent=1)
    elif filename == "STDOUT":
        sys.stdout.write(
            ujson.dumps(data, indent=1, escape_forward_slashes=False) + "\n"
        )
        return True
    elif filename == "STDOUT":
        sys.stderr.write(
            ujson.dumps(data, indent=1, escape_forward_slashes=False) + "\n"
        )
        return True
    elif plain:
        content = "\n".join(data)
    elif ".csv" in filename or ".tsv" in filename:
        output = io.StringIO()
        if ".csv" in filename:
            writer = csv.writer(output, quoting=csv.QUOTE_NONNUMERIC)
        else:
            writer = csv.writer(output, delimiter="\t")
        for row in data:
            writer.writerow(row)
        content = output.getvalue()
    else:
        content = data
    if ".gz" in filename:
        try:
            with gzip.open(filename, "wt") as fh:
                fh.write(content)
        except TypeError:
            with open(filename, "wb") as fh:
                fh.write(content)
        except OSError:
            return False
    else:
        try:
            with open(filename, "wt") as fh:
                fh.write(content)
        except TypeError:
            with open(filename, "wb") as fh:
                fh.write(content)
        except IOError:
            return False
    return True


def delete_file(filename):
    """
    Delete a file if exists.

    Args:
        filename (str): Name of file to delete.
    """
    if os.path.exists(filename):
        os.remove(filename)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python3

"""Binning functions."""

import math


def readable_bin(value, *, start_digits=None):
    """
    Place values in human readable bins.

    Args:
        value (float): Number to bin.
        start_digits (list, optional): List of threshold values between 1 and 10.
            Defaults to `[1, 2, 3, 5]`

    .. testsetup::

       from tolkein.lib.tobin import readable_bin

    .. doctest::

        >>> readable_bin(123456789)
        '200M'

        >>> readable_bin(56789012234)
        '100G'

        >>> readable_bin(567.89)
        '1k'

        >>> readable_bin(21, start_digits=[1, 2, 4, 8])
        '40'

        >>> readable_bin(-1234567)
        '-2M'
    """
    if start_digits is None:
        start_digits = [1, 2, 3, 5, 10]
    else:
        start_digits.append(start_digits[0] * 10)
    thousands = ["", "k", "M", "G", "T", "P", "E", "Z"]
    start_digits.append(10)
    # Check sign
    sign = 1
    if value < 0:
        value = abs(value)
        sign = -1
    # Check order of magnitude
    order = int(math.floor(math.log10(abs(value))))
    # Get first digit
    first = value / pow(10, order)
    # Find bin
    for digit in start_digits:
        if first < digit:
            initial = digit
            if initial == 10:
                order += 1
                initial = 1
            break
    # Count thousands
    mag = order // 3
    # Count additional orders of magnitude
    order %= 3
    # Set return value
    value = sign * initial * pow(10, order)
    return "%d%s" % (value, thousands[mag])


if __name__ == "__main__":
    # Run inline tests
    import doctest

    doctest.testmod()



#!/usr/bin/env python3

"""Fetch methods."""

import os
import shutil
import tarfile
import urllib.request as request
import zlib
from contextlib import closing
from typing import Union
from urllib.error import URLError

import requests
from tqdm import tqdm

from .tofile import write_file

url = "ftp://ftp.myurl.com"


class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`.

    From tqdm documentation.
    """

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        Tqdm update_to method.

        Args:
            b (int, optional): Number of blocks transferred so far [default: 1].
            bsize (int, optional): Size of each block (in tqdm units) [default: 1].
            tsize (int, optional): Total size (in tqdm units).
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def fetch_tmp_file(url):
    """
    Fetch a remote URL to a temporary file.

    Args:
        url (str) : Remote URL to fetch.

    Returns:
        str : Temporary filename.
    """
    with TqdmUpTo(
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        miniters=1,
        desc="Fetch %s" % url.split("/")[-1],
    ) as t:
        file_tmp = request.urlretrieve(
            url, filename=None, reporthook=t.update_to, data=None
        )[0]
        t.total = t.n
    return file_tmp


def extract_tar(filename, path):
    """
    Extract tarred archive.

    Args:
        filename (str) : Name of tar file to extract.
        path (str) : Path to extract tar file.

    Returns:
        int : Count of file members extracted from tar archive.
    """
    member_count = 0
    with tarfile.open(filename) as tar:
        member_count = len(tar.getmembers())
        # Go over each member
        for member in tqdm(
            iterable=tar.getmembers(), total=member_count, desc="Extract files"
        ):
            # Extract member
            tar.extract(member=member, path=path)
    return member_count


def fetch_tar(url, path):
    """
    Fetch and extract tarred archives.

    Args:
        url (str): Remote URL to fetch.
        path (str): Path to extract tar file.

    Returns:
        int: Count of file members extracted from tar archive.
    """
    file_tmp = fetch_tmp_file(url)
    member_count = extract_tar(file_tmp, path)
    return member_count


def fetch_stream(url, *, decode=True, show_progress=True):
    """
    Stream download.

    Args:
        decode (bool, optional): Determines whether to unzip content. Defaults to True.
        show_progress (bool, optional): Show a progress bar to indicate file streaming
            progress. Defaults to True.

    Yields:
        str: 1024 byte chunk of remote URL.
    """
    res = requests.get(url, stream=True)
    if res.encoding is None:
        res.encoding = "utf-8"
    total_size = int(res.headers.get("content-length", 0))
    block_size = 1024
    if show_progress:
        progress = tqdm(total=total_size, unit="iB", unit_scale=True)
    dec = zlib.decompressobj(32 + zlib.MAX_WBITS)
    for data in res.iter_content(block_size):
        if decode:
            try:
                data = dec.decompress(data)
                if show_progress:
                    progress.update(len(data))
                yield data
            except zlib.error:
                decode = False
        if show_progress:
            progress.update(len(data))
        yield data
    if show_progress:
        progress.close()


def fetch_file(url, path, decode=True):
    """Fetch a remote file.

    Args:
        url (str): Remote URL to fetch.
        path (str): Path to extract tar file.
        decode (bool, optional): Determines whether to unzip content. Defaults to True.
    """
    data = "" if decode else b""
    for part in fetch_stream(url, decode=decode):
        data += part.decode("utf-8") if decode else part
    write_file(path, data)


def fetch_url(url):
    """Fetch a URL.

    Args:
        url (str): Remote URL to fetch.

    Returns:
        str: Content of file as a string. Will return None if response is not OK.
    """
    res = requests.get(url)
    if res.ok:
        return res.content.decode("utf-8")
    return None


def fetch_ftp(url, filename):
    """Fetch a file via ftp."""
    try:
        with closing(request.urlopen(url)) as req:
            with open(filename, "wb") as fh:
                shutil.copyfileobj(req, fh)
    except URLError:
        return None



#!/usr/bin/env python3

"""
Fill attribute values.

Usage:
    genomehubs fill [--hub-name STRING] [--hub-path PATH] [--hub-version PATH]
                    [--config-file PATH...] [--config-save PATH]
                    [--es-batch INT] [--es-host URL...]  [--taxonomy-source STRING]
                    [--traverse-limit STRING]
                    [--traverse-infer-ancestors] [--traverse-infer-descendants]
                    [--traverse-infer-both] [--traverse-threads INT]
                    [--traverse-depth INT] [--traverse-root STRING]
                    [--traverse-weight STRING] [--log-interval INT]
                    [--log-es BOOL]
                    [-h|--help] [-v|--version]

Options:
    --hub-name STRING             GenomeHubs instance name.
    --hub-path PATH               GenomeHubs instance root directory path.
    --hub-version STR             GenomeHubs instance version string.
    --config-file PATH            Path to YAML file containing configuration options.
    --config-save PATH            Path to write configuration options to YAML file.
    --es-batch INT                Batch size for ElasticSearch bulk indexing.
    --es-host URL                 ElasticSearch hostname/URL and port.
    --taxonomy-source STRING      Name of taxonomy to use (ncbi or ott).
    --traverse-depth INT          Maximum depth for tree traversal relative to root taxon.
    --traverse-infer-ancestors    Flag to enable tree traversal from tips to root.
    --traverse-infer-descendants  Flag to enable tree traversal from root to tips.
    --traverse-infer-both         Flag to enable tree traversal from tips to root and
                                  back to tips.
    --traverse-limit STRING       Maximum rank to ascend to during traversal. [Default: null]
    --traverse-root ID            Root taxon id for tree traversal.
    --traverse-threads INT        Number of threads to use for tree traversal. [Default: 1]
    --traverse-weight STRING      Weighting scheme for setting values during tree
                                  traversal.
    --log-interval INT            Minimum time (seconds) between prgress bar updates.
    --log-es BOOL                 Show Info-level logs from elasticsearch.
    -h, --help                    Show this
    -v, --version                 Show version number

Examples:
    # 1. Traverse tree up to taxon_id 7088
    ./genomehubs fill --traverse-root 7088
"""


import contextlib
import re
import sys
from collections import defaultdict
from datetime import datetime
from itertools import groupby
from multiprocessing import Pool
from statistics import mean
from statistics import median
from statistics import median_high
from statistics import median_low
from statistics import mode
from traceback import format_exc

from docopt import docopt
from h3 import latlng_to_cell
from tqdm import tqdm

from genomehubs.vendor.tolkein import tolog

from ..lib import hub
from ..lib import taxon
from .attributes import fetch_types
from .config import config
from .es_functions import index_stream
from .es_functions import launch_es
from .es_functions import stream_template_search_results
from .version import __version__

LOGGER = tolog.logger(__name__)


def get_max_depth(es, *, index):
    """Find max depth of root lineage."""
    body = {
        "id": "max_nested_value",
        "params": {"path": "lineage", "field": "node_depth"},
    }
    res = es.search_template(index=index, body=body)
    return res["aggregations"]["depths"]["max_depth"]["value"]


def get_max_depth_by_lineage(es, *, index, root):
    """Find max depth of specified root lineage."""
    body = {
        "id": "max_nested_value_by_key_value",
        "params": {
            "path": "lineage",
            "key": "taxon_id",
            "value": root,
            "field": "node_depth",
        },
    }
    res = es.search_template(index=index, body=body)
    return res["aggregations"]["depths"]["root"]["max_depth"]["value"]


def stream_nodes_by_root_depth(es, *, index, root, depth, size=10):
    """Get entries by depth of root taxon."""
    if depth > 0:
        body = {
            "id": "taxon_attributes_by_root_depth",
            "params": {"taxon_id": root, "depth": depth},
        }
        return stream_template_search_results(es, index=index, body=body, size=size)
    body = {
        "id": "taxon_attributes_by_taxon_id",
        "params": {"taxon_id": root},
    }
    return stream_template_search_results(es, index=index, body=body)


def stream_descendant_nodes_missing_attributes(es, *, index, attributes, root, size=10):
    """Get entries descended from root that lack one or more attributes."""
    id_list = set()
    for attribute in attributes:
        body = {
            "id": "taxon_missing_attribute_by_ancestor_id",
            "params": {"taxon_id": root, "attribute": attribute["key"]},
        }
        for result in stream_template_search_results(
            es, index=index, body=body, size=size
        ):
            if result["_id"] not in id_list:
                yield result


def enum(tup):
    """Use list index to prioritise values."""
    order = {str(k).lower(): i for i, k in enumerate(tup[0])}
    lowest = len(tup[0]) + 1
    result = None
    for value in tup[1]:
        lc_value = str(value).lower()
        if lc_value in order and order[lc_value] < lowest:
            result = value
            lowest = order[lc_value]
            if lowest == 0:
                break
    return result


def ordered_list(tup):
    """Remove values that are in a higher priority list."""
    (key, order, arr, linked) = tup
    values = deduped_list(arr)
    seen = set()
    for i, k in enumerate(order):
        if i == 0 and k == key:
            break
        if k == key:
            return [value for value in values if value not in seen]
        if k in linked:
            with contextlib.suppress(KeyError):
                seen.update(linked[k]["keyword_value"])
    return values


def earliest(arr, *args):
    """Select earliest date from a list."""
    if not isinstance(arr, list):
        arr = [arr]
    if arr and isinstance(arr[0], str):
        date_arr = [datetime.strptime(date.split("T")[0], "%Y-%m-%d") for date in arr]
        return datetime.strftime(min(date_arr), "%Y-%m-%d")
    return min(arr + list(args))


def latest(arr, *args):
    """Select earliest date from a list."""
    if not isinstance(arr, list):
        arr = [arr]
    if isinstance(arr[0], str) and re.match(r"\d{4}-\d{2}-\d{2}", arr[0]):
        date_arr = [datetime.strptime(date.split("T")[0], "%Y-%m-%d") for date in arr]
        return datetime.strftime(max(date_arr), "%Y-%m-%d")
    return max(arr + list(args))


def range(arr):
    """Calculate difference between max and min values."""
    return latest(arr) - earliest(arr)


def median_list(arr):
    """Return both values in event of tied median."""
    length = len(arr)
    if length % 2 == 1:
        return [median(arr)]
    sorted_arr = sorted(arr, reverse=True)
    return list(set([sorted_arr[int(length / 2)], sorted_arr[int(length / 2) - 1]]))


def mode_list(arr):
    """Return a list of modal values."""
    mode_count = 0
    mode_arr = []
    for key, group in groupby(arr):
        count = len(list(group))
        if count < mode_count:
            continue
        if count > mode_count:
            mode_count = count
            mode_arr = []
        mode_arr.append(key)
    return mode_arr


def mode_high(arr):
    """Calculate mode using median_high to resolve ties."""
    return median_high(mode_list(arr))


def mode_low(arr):
    """Calculate mode using median_low to resolve ties."""
    return median_low(mode_list(arr))


def mode_mean(arr):
    """Calculate mode using mean to resolve ties."""
    return mean(mode_list(arr))


def flatten_list(arr):
    """Flatten a list by expanding any nested lists."""
    flattened = []
    for v in arr:
        if isinstance(v, list):
            flattened += v
        else:
            flattened.append(v)
    return flattened


def deduped_list(arr):
    """Remove duplicate values from a list."""
    return list(set(flatten_list(arr)))


def hex_bin(arr, resolution=6):
    """Convert list of point coordinates to h3 hexagons."""
    hexagons = []
    for coord in deduped_list(arr):
        if isinstance(coord, str):
            # Handle comma-separated string "lat,lon"
            lat_str, lon_str = coord.split(",")
            lat, lon = float(lat_str.strip()), float(lon_str.strip())
            hexagons.append(latlng_to_cell(lat, lon, resolution))
        elif isinstance(coord, list):
            # Handle list of [lat, lon] or list of such lists/strings
            for item in coord:
                if isinstance(item, str):
                    lat_str, lon_str = item.split(",")
                    lat, lon = float(lat_str.strip()), float(lon_str.strip())
                    hexagons.append(latlng_to_cell(lat, lon, resolution))
                elif isinstance(item, (list, tuple)) and len(item) == 2:
                    lat, lon = float(item[0]), float(item[1])
                    hexagons.append(latlng_to_cell(lat, lon, resolution))
        elif isinstance(coord, (tuple, list)) and len(coord) == 2:
            lat, lon = float(coord[0]), float(coord[1])
            hexagons.append(latlng_to_cell(lat, lon, resolution))
    return deduped_list(hexagons)


def hexbin1(arr):
    """Convert list of coordinates to hexagonal bins at resolution 1."""
    return hex_bin(arr, resolution=1)


def hexbin2(arr):
    """Convert list of coordinates to hexagonal bins at resolution 2."""
    return hex_bin(arr, resolution=2)


def hexbin3(arr):
    """Convert list of coordinates to hexagonal bins at resolution 3."""
    return hex_bin(arr, resolution=3)


def hexbin4(arr):
    """Convert list of coordinates to hexagonal bins at resolution 4."""
    return hex_bin(arr, resolution=4)


def hexbin5(arr):
    """Convert list of coordinates to hexagonal bins at resolution 5."""
    return hex_bin(arr, resolution=5)


def hexbin6(arr):
    """Convert list of coordinates to hexagonal bins at resolution 6."""
    return hex_bin(arr, resolution=6)


def deduped_list_length(arr):
    """Find number of unique values in a list."""
    return len(deduped_list(arr))


def apply_summary(
    summary,
    values,
    *,
    primary_values=None,
    summary_types=None,
    max_value=None,
    min_value=None,
    order=None,
    attr_order=None,
    meta=None,
    linked_attributes=None,
):
    """Apply summary statistic functions."""
    summaries = {
        "count": len,
        "earliest": earliest,
        "enum": enum,
        "latest": latest,
        "max": latest,
        "min": earliest,
        "mean": mean,
        "median": median,
        "median_high": median_high,
        "median_low": median_low,
        "median_list": median_list,
        "mode": mode,
        "most_common": mode,
        "mode_high": mode_high,
        "mode_low": mode_low,
        "mode_mean": mode_mean,
        "mode_list": mode_list,
        "range": range,
        "sum": sum,
        "sp_count": len,
        "list": deduped_list,
        "length": deduped_list_length,
        "ordered_list": ordered_list,
        "hexbin1": hexbin1,
        "hexbin2": hexbin2,
        "hexbin3": hexbin3,
        "hexbin4": hexbin4,
        "hexbin5": hexbin5,
        "hexbin6": hexbin6,
    }
    if summary == "primary":
        if primary_values:
            values = primary_values
        summary = summary_types[0]
    flattened = flatten_list(values)
    if summary == "enum":
        value = summaries[summary]((order, flattened))
    elif summary == "ordered_list":
        value = summaries[summary](
            (meta["key"], attr_order, flattened, linked_attributes)
        )
    else:
        value = summaries[summary](flattened)
    if summary == "max":
        if max_value is not None:
            value = latest(value, max_value)
        max_value = value
    elif summary == "min":
        if min_value is not None:
            value = earliest(value, min_value)
        min_value = value
    return value, max_value, min_value


def set_traverse_values(
    summaries,
    values,
    primary_values,
    count,
    sp_count,
    max_value,
    min_value,
    meta,
    attribute,
    value_type,
    traverse,
    source,
    linked_attributes,
):
    """Set values  use for tree traversal."""
    idx = 0
    order = meta.get("constraint", {}).get("enum", [])
    attr_order = meta.get("order", [])
    default_summary = "median"
    if meta["type"] == "keyword":
        default_summary = "mode"
    elif meta["type"] == "date":
        default_summary = "latest"
    for index, summary in enumerate(summaries):
        summary_types = meta["summary"][index + 1 :] + [default_summary]
        value, max_value, min_value = apply_summary(
            summary,
            values,
            primary_values=primary_values,
            summary_types=summary_types,
            max_value=max_value,
            min_value=min_value,
            order=order,
            attr_order=attr_order,
            meta=meta,
            linked_attributes=linked_attributes,
        )
        if idx == 0:
            if value is not None:
                if (
                    value_type not in attribute
                    or attribute["aggregation_method"] != "primary"
                ):
                    if summary == "primary" and "values" not in attribute:
                        summary = summary_types[0]
                    attribute[value_type] = value
                    attribute["count"] = count or len(values)
                    attribute["sp_count"] = sp_count
                    if summary in ["list", "ordered_list"]:
                        attribute["length"] = deduped_list_length(values)
                    attribute["aggregation_method"] = summary
                    attribute["aggregation_source"] = source
                traverse_value = value or []
            idx += 1
        if traverse and source == "descendant" and summary == traverse:
            traverse_value = value or []
        elif summary != "list":
            if summary.startswith("median"):
                summary = "median"
        else:
            traverse_value = list(set(traverse_value))
        if summary == "range":
            attribute[summary] = value
            traverse_value = [min_value, max_value]
        elif summary not in attribute and summary in {
            "mean",
            "median",
            "mode",
            "sum",
            "hexbin1",
            "hexbin2",
            "hexbin3",
            "hexbin4",
            "hexbin5",
            "hexbin6",
        }:
            attribute[summary] = value
    return traverse_value, max_value, min_value


# def iterate_values(attribute, meta):
#     """Iterate through values adding prefixed values if appropriate."""
#     prefixed_values = None
#     if meta["type"] == "keyword":
#         prefixed_values = attribute.get("prefixed_value", [])
#     value_type = "%s_value" % meta["type"]
#     for value in attribute["values"]:
#         if prefixed_values is not None and "source_prefix" in value:
#             prefixed_values.append(
#                 "%s:%s" % (value["source_prefix"], value[value_type])
#             )
#     if prefixed_values:
#         attribute["prefixed_values"] = prefixed_values


def summarise_attribute_values(
    attribute,
    meta,
    *,
    linked_attributes=None,
    values=None,
    count=0,
    sp_count=0,
    max_value=None,
    min_value=None,
    source="direct",
):
    """Calculate a single summary value for an attribute."""
    if values is None and "values" not in attribute:
        return None, None, None
    if "summary" in meta:
        value_type = f'{meta["type"]}_value'
        primary_values = []
        if "values" in attribute:
            # iterate_values(attribute, meta)
            if values is None:
                values = []
                for value in attribute["values"]:
                    try:
                        values.append(value[value_type])
                    except KeyError:
                        print(meta)
                        print(value)
                    if "is_primary_value" in value and value["is_primary_value"]:
                        primary_values.append(value[value_type])
            else:
                # TODO: handle existing value here
                values += [value[value_type] for value in attribute["values"]]
        if not values:
            return None, None, None
        traverse = meta.get("traverse", False)
        traverse_value = None
        if not isinstance(meta["summary"], list):
            meta["summary"] = [meta["summary"]]
        try:
            summaries = meta["summary"][:]
            if traverse and source != "ancestor" and summaries[0] != traverse:
                if summaries[0] != "primary":
                    summaries = [traverse] + summaries
                elif summaries[1] != "primary":
                    summaries.insert(1, traverse)
            traverse_value, max_value, min_value = set_traverse_values(
                summaries,
                values,
                primary_values,
                count,
                sp_count,
                max_value,
                min_value,
                meta,
                attribute,
                value_type,
                traverse,
                source,
                linked_attributes,
            )
        except Exception:
            print(format_exc())
            LOGGER.error(
                f"Unable to generate summary values for attribute {meta['key']}"
            )
            sys.exit(1)
        if isinstance(max_value, (float, int)):
            attribute["max"] = max_value
            attribute["min"] = min_value
        elif meta["type"] == "date" and max_value and min_value:
            attribute["from"] = min_value
            attribute["to"] = max_value
        return traverse_value, max_value, min_value
    return None, None, None


def summarise_attributes(*, attributes, rank, attrs, meta, parent, parents):
    """Set attribute summary values."""
    changed = False
    attr_dict = {}
    for node_attribute in attributes:
        if node_attribute["key"] in attrs:
            attr_dict[node_attribute["key"]] = node_attribute
            if rank == "species":
                parents[parent][node_attribute["key"]]["sp_count"] += 1
                attr_dict[node_attribute["key"]]["sp_count"] = 1
            elif rank in ["subspecies", "varietas", "strain"]:
                parents[parent][node_attribute["key"]]["sp_count"] = 1
                attr_dict[node_attribute["key"]]["sp_count"] = 0

            linked_attributes = {}
            if "order" in meta[node_attribute["key"]]:
                for attribute in attributes:
                    if attribute["key"] in meta[node_attribute["key"]]["order"]:
                        linked_attributes[attribute["key"]] = attribute
            summary_value, max_value, min_value = summarise_attribute_values(
                node_attribute,
                {"key": node_attribute["key"], **meta[node_attribute["key"]]},
                linked_attributes=linked_attributes,
                # sp_count=sp_count,
            )
            if summary_value is not None:
                changed = True
                if parent is not None:
                    parents[parent][node_attribute["key"]]["count"] += 1
                    if isinstance(summary_value, list):
                        parents[parent][node_attribute["key"]][
                            "values"
                        ] += summary_value
                    else:
                        parents[parent][node_attribute["key"]]["values"].append(
                            summary_value
                        )
                    if max_value is not None:
                        if parents[parent][node_attribute["key"]]["max"] is not None:
                            parents[parent][node_attribute["key"]]["max"] = latest(
                                parents[parent][node_attribute["key"]]["max"], max_value
                            )
                        else:
                            parents[parent][node_attribute["key"]]["max"] = max_value
                    if min_value is not None:
                        if parents[parent][node_attribute["key"]]["min"] is not None:
                            parents[parent][node_attribute["key"]]["min"] = earliest(
                                parents[parent][node_attribute["key"]]["min"], min_value
                            )
                        else:
                            parents[parent][node_attribute["key"]]["min"] = min_value
    return changed, attr_dict


def set_aggregation_source(attribute, source=None):
    """Set attribute aggregation source."""
    if "aggregation_source" not in attribute:
        return_value = "descendant"
    else:
        return_value = attribute["aggregation_source"]
    if source is not None:
        if "aggregation_source" not in attribute:
            attribute["aggregation_source"] = source
        elif attribute["aggregation_source"] == "direct":
            attribute["aggregation_source"] = ["direct", source]
    return return_value


def set_values_from_descendants(
    *,
    attributes,
    descendant_values,
    meta,
    taxon_id,
    parent,
    taxon_rank,
    traverse_limit,
    parents,
    descendant_ranks=None,
    attr_dict=None,
    limits=None,
):
    """Set attribute summary values from descendant values."""
    changed = False
    if attr_dict is None:
        attr_dict = {}
    for key, obj in descendant_values.items():
        traverseable = meta[key].get("traverse", False)
        if (
            traverseable
            and "traverse_direction" in meta[key]
            and meta[key]["traverse_direction"] == "down"
        ):
            traverseable = False
        if not traverseable or taxon_id in limits[key]:
            continue
        if local_limit := meta[key].get("traverse_limit", traverse_limit):
            if (
                descendant_ranks is not None
                and local_limit in descendant_ranks[taxon_id]
            ):
                continue
            if taxon_rank == local_limit:
                limits[key].add(parent)
        try:
            attribute = next(entry for entry in attributes if entry["key"] == key)
        except StopIteration:
            attribute = {"key": key}
            attributes.append(attribute)
        linked_attributes = {}
        if "order" in meta[key]:
            for attribute in attributes:
                if attribute["key"] in meta[key]["order"]:
                    linked_attributes[attribute["key"]] = attribute
        summary_value, max_value, min_value = summarise_attribute_values(
            attribute,
            {"key": key, **meta[key]},
            values=obj["values"],
            count=obj["count"],
            sp_count=obj["sp_count"],
            max_value=obj["max"],
            min_value=obj["min"],
            source=set_aggregation_source(attribute),
            linked_attributes=linked_attributes,
        )
        set_aggregation_source(attribute, "descendant")

        if summary_value is not None:
            changed = True
            attr_dict.update({key: attribute})
            if parent is not None:
                parents[parent][key]["count"] += 1
                if "sp_count" not in parents[parent][key]:
                    parents[parent][key]["sp_count"] = 0
                if "sp_count" in attribute:
                    parents[parent][key]["sp_count"] += attribute["sp_count"]
                if isinstance(summary_value, list):
                    parents[parent][key]["values"] = list(
                        set(parents[parent][key]["values"] + summary_value)
                    )
                else:
                    parents[parent][key]["values"].append(summary_value)
                if max_value is not None:
                    if parents[parent][key]["max"] is not None:
                        parents[parent][key]["max"] = latest(
                            parents[parent][key]["max"], max_value
                        )
                    else:
                        parents[parent][key]["max"] = max_value
                if min_value is not None:
                    if parents[parent][key]["min"] is not None:
                        parents[parent][key]["min"] = earliest(
                            parents[parent][key]["min"], min_value
                        )
                    else:
                        parents[parent][key]["min"] = min_value
    return changed, attr_dict


def set_attributes_to_descend(meta, traverse_limit):
    """Set which attributes should have values inferred from ancestral taxa."""
    desc_attrs = set()
    desc_attr_limits = {}
    for key, value in meta.items():
        if (
            "traverse" in value
            and value["traverse"]
            and (
                "traverse_direction" not in value
                or value["traverse_direction"]
                in (
                    "down",
                    "both",
                )
            )
        ):
            desc_attrs.add(key)
            if "traverse_limit" in value:
                desc_attr_limits[key] = value["traverse_limit"]
            elif traverse_limit != "null":
                desc_attr_limits[key] = traverse_limit
    return desc_attrs, desc_attr_limits


def track_missing_attribute_values(
    node, missing_attributes, attr_dict, desc_attrs, desc_attr_limits
):
    """Keep track of missing attribute values for in memory traversal."""
    missing_from_descendants = {}
    if (
        node["_source"]["taxon_id"] in missing_attributes
        and missing_attributes[node["_source"]["taxon_id"]]
    ):
        for child_id, obj in missing_attributes[node["_source"]["taxon_id"]].items():
            for key, attribute in attr_dict.items():
                if key in obj["keys"]:
                    # update aggregation source to include ancestral rank
                    obj["attributes"].append(
                        {
                            **attribute,
                            "aggregation_source": "ancestor",
                            "aggregation_rank": node["_source"]["taxon_rank"],
                            "aggregation_taxon_id": node["_source"]["taxon_id"],
                        }
                    )
                    obj["keys"].remove(key)
                    obj["attributes"][-1].pop("sp_count", None)
            if obj["keys"]:
                missing_from_descendants[child_id] = obj
            else:
                # yield when all values filled or removed
                yield obj["node"]["_id"], obj["node"]["_source"]
        del missing_attributes[node["_source"]["taxon_id"]]
    if "parent" in node["_source"]:
        missing_attributes[node["_source"]["parent"]].update(missing_from_descendants)
        missing_attributes[node["_source"]["parent"]].update(
            {
                node["_source"]["taxon_id"]: {
                    "keys": set({key for key in desc_attrs if key not in attr_dict}),
                    "attributes": node["_source"]["attributes"],
                    "node": node,
                }
            }
        )


def track_descendant_ranks(node, descendant_ranks):
    """Keep track of descendant ranks."""
    if "parent" in node["_source"]:
        descendant_ranks[node["_source"]["parent"]].add(node["_source"]["taxon_rank"])


def traverse_from_tips(es, opts, *, template, root=None, max_depth=None):
    """Traverse a tree, filling in values."""
    if root is None:
        root = opts["traverse-root"]
    if max_depth is None:
        max_depth = get_max_depth_by_lineage(
            es,
            index=template["index_name"],
            root=root,
        )
    root_depth = max_depth
    meta = template["types"]["attributes"]
    attrs = set(meta.keys())
    parents = defaultdict(
        lambda: defaultdict(
            lambda: {
                "max": None,
                "min": None,
                "values": [],
                "prefixed_values": [],
                "count": 0,
                "sp_count": 0,
            }
        )
    )
    limits = defaultdict(set)
    if "traverse-infer-both" in opts and opts["traverse-infer-both"]:
        desc_attrs, desc_attr_limits = set_attributes_to_descend(
            meta, opts["traverse-limit"]
        )
        missing_attributes = defaultdict(dict)
        descendant_ranks = defaultdict(set)
    else:
        desc_attrs = {}
    while root_depth >= 0:
        nodes = stream_nodes_by_root_depth(
            es,
            index=template["index_name"],
            root=root,
            depth=root_depth,
            size=50,
        )
        for ctr, node in enumerate(nodes):
            track_descendant_ranks(node, descendant_ranks)
            changed = False
            attr_dict = {}
            if "attributes" in node["_source"] and node["_source"]["attributes"]:
                changed, attr_dict = summarise_attributes(
                    attributes=node["_source"]["attributes"],
                    rank=node["_source"]["taxon_rank"],
                    attrs=attrs,
                    meta=meta,
                    parent=node["_source"].get("parent", None),
                    parents=parents,
                )
            else:
                node["_source"]["attributes"] = []
            if node["_source"]["taxon_id"] in parents:
                modified, attr_dict = set_values_from_descendants(
                    attributes=node["_source"]["attributes"],
                    descendant_values=parents[node["_source"]["taxon_id"]],
                    meta=meta,
                    taxon_id=node["_source"]["taxon_id"],
                    parent=node["_source"].get("parent", None),
                    parents=parents,
                    descendant_ranks=descendant_ranks,
                    taxon_rank=node["_source"]["taxon_rank"],
                    traverse_limit=opts["traverse-limit"],
                    attr_dict=attr_dict,
                    limits=limits,
                )
                if not changed:
                    changed = modified
            if desc_attrs:
                yield from track_missing_attribute_values(
                    node, missing_attributes, attr_dict, desc_attrs, desc_attr_limits
                )
            if changed:
                yield node["_id"], node["_source"]
        root_depth -= 1
    if desc_attrs:
        for incomplete in missing_attributes.values():
            for obj in incomplete.values():
                yield obj["node"]["_id"], obj["node"]["_source"]


def copy_attribute_summary(source, meta):
    """Copy an attribute summary, removing values."""
    dest = {}
    for key in meta["summary"]:
        if key.startswith("median") and "median" in source:
            dest["median"] = source["median"]
        elif key != "list" and key in source:
            dest[key] = source[key]
    try:
        dest[f'{meta["type"]}_value'] = source[f'{meta["type"]}_value']
    except KeyError as err:
        raise (err)
    dest["count"] = source["count"]
    dest["sp_count"] = source.get("sp_count", 0)
    dest["key"] = source["key"]
    return dest


def stream_missing_attributes_at_level(es, *, nodes, attrs, template, level=1):
    """Stream all descendant nodes with missing attributes."""
    for node in nodes:
        taxon_id = node["_source"]["taxon_id"]
        fill_attrs = []
        if "attributes" in node["_source"]:
            fill_attrs.extend(
                attribute
                for attribute in node["_source"]["attributes"]
                if attribute["key"] in attrs
            )
        if not fill_attrs:
            continue
        meta = template["types"]["attributes"]
        desc_nodes = stream_descendant_nodes_missing_attributes(
            es,
            index=template["index_name"],
            attributes=fill_attrs,
            root=taxon_id,
            size=50,
        )
        anc_attributes = {attr["key"]: attr for attr in fill_attrs}
        for desc_node in desc_nodes:
            skip_attrs = set()
            if "attributes" in desc_node["_source"]:
                for desc_attribute in desc_node["_source"]["attributes"]:
                    if desc_attribute["key"] in anc_attributes:
                        skip_attrs.add(desc_attribute["key"])
            else:
                desc_node["_source"]["attributes"] = []
            for key, anc_attribute in anc_attributes.items():
                if key not in skip_attrs:
                    desc_attribute = copy_attribute_summary(anc_attribute, meta[key])
                    desc_attribute["aggregation_method"] = meta[key]["traverse"]
                    desc_attribute["aggregation_source"] = "ancestor"
                    desc_node["_source"]["attributes"].append(desc_attribute)
            yield desc_node["_id"], desc_node["_source"]


def traverse_from_root(es, opts, *, template, root=None, max_depth=None, log=True):
    """Traverse a tree, filling in values."""
    if root is None:
        root = opts["traverse-root"]
    if max_depth is None:
        max_depth = get_max_depth_by_lineage(
            es, index=template["index_name"], root=root
        )
    root_depth = max_depth - 1
    meta = template["types"]["attributes"]
    attrs = set({})
    for key, value in meta.items():
        if (
            "traverse" in value
            and value["traverse"]
            and (
                "traverse_direction" not in value or value["traverse_direction"] != "up"
            )
        ):
            attrs.add(key)
    while root_depth >= 0:
        LOGGER.info("Filling values at root depth %d" % root_depth)
        nodes = stream_nodes_by_root_depth(
            es, index=template["index_name"], root=root, depth=root_depth, size=50
        )
        desc_nodes = stream_missing_attributes_at_level(
            es, nodes=nodes, attrs=attrs, template=template
        )
        index_stream(
            es,
            template["index_name"],
            desc_nodes,
            _op_type="update",
            log=opts.get("log-es", True),
            chunk_size=opts.get("es-batch", 500),
        )
        root_depth -= 1


def traverse_tree(es, opts, template, root, max_depth):
    """Propagate values by tree traversal."""
    log = True
    if es is None:
        log = False
        es = launch_es(opts, log=log)
    if "traverse-infer-ancestors" in opts:
        LOGGER.info("Inferring ancestral values for root taxon %s", root)
        _success, _failed = index_stream(
            es,
            template["index_name"],
            traverse_from_tips(
                es,
                opts,
                template=template,
                root=root,
                max_depth=max_depth,
            ),
            _op_type="update",
            log=opts.get("log-es", True),
            chunk_size=opts.get("es-batch", 500),
        )
    if "traverse-infer-descendants" in opts:
        if log:
            LOGGER.info("Inferring descendant values for root taxon %s", root)
        traverse_from_root(
            es, opts, template=template, root=root, max_depth=max_depth, log=log
        )


def traverse_helper(params):
    """Wrap traverse_tree for multithreaded traversal."""
    with tolog.DisableLogger():
        traverse_tree(*params)
    return params[3]


def traverse_handler(es, opts, template):
    """Handle single or multi-threaded tree traversal."""
    root = opts["traverse-root"]
    threads = int(opts["traverse-threads"])
    max_depth = get_max_depth_by_lineage(es, index=template["index_name"], root=root)
    subtree_depth = int(opts.get("traverse-depth", 0))
    if threads > 1 and subtree_depth == 0:
        subtree_depth = int(max_depth / 2)
    if threads == 1:
        traverse_tree(es, opts, template, root, max_depth)
        return

    max_depth -= subtree_depth

    nodes = stream_nodes_by_root_depth(
        es, index=template["index_name"], root=root, depth=subtree_depth
    )
    roots = [
        (None, opts, template, node["_source"]["taxon_id"], max_depth) for node in nodes
    ]
    LOGGER.info("Filling values in subtrees")
    with Pool(processes=threads) as p:
        with tqdm(
            total=len(roots),
            unit=" subtrees",
            mininterval=int(opts.get("log-interval", 1)),
        ) as pbar:
            for _ in p.imap_unordered(traverse_helper, roots):
                pbar.update()
    LOGGER.info("Connecting subtrees")
    traverse_tree(es, opts, template, opts["traverse-root"], subtree_depth)


def main(args):
    """Initialise genomehubs."""
    options = config("fill", **args)
    if "traverse-infer-both" in options["fill"]:
        options["fill"]["traverse-infer-ancestors"] = True

    # Start Elasticsearch
    es = launch_es(options["fill"])

    # Post search scripts
    hub.post_search_scripts(es)

    LOGGER.info("Filling values")

    types = fetch_types(es, "taxon", options["fill"])

    if "taxonomy-source" in options["fill"]:
        taxonomy_name = options["fill"]["taxonomy-source"].lower()
        template = taxon.index_template(taxonomy_name, options["fill"])
        if types:
            template["types"]["attributes"] = types
        if "traverse-root" in options["fill"]:
            traverse_handler(es, options["fill"], template)


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index(__name__.split(".")[-1]) + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()
    cli()



#!/usr/bin/env python3
"""Taxonomy methods."""

import re

from .tofile import open_file_handle


def parse_ncbi_nodes_dmp(path):
    """Parse NCBI format nodes.dmp file."""
    nodes = {}
    with open_file_handle(path) as fh:
        for line in fh:
            taxon_id, parent, taxon_rank, *_ignore = re.split(r"\s*\|\s*", line)
            nodes[taxon_id] = {
                "parent": parent,
                "taxon_rank": taxon_rank,
                "taxon_names": [],
            }
        nodes["1"] = {"taxon_rank": "no rank", "taxon_names": []}
    return nodes


def parse_ncbi_names_dmp(path, nodes):
    """Parse names.dmp file and add to nodes dict."""
    with open_file_handle(path) as fh:
        for line in fh:
            taxon_id, name, _unique, name_class, *_ignore = re.split(r"\s*\|\s*", line)
            if taxon_id in nodes:
                if name_class == "scientific name":
                    nodes[taxon_id].update(
                        {
                            "taxon_id": taxon_id,
                            "scientific_name": name,
                        }
                    )
                nodes[taxon_id]["taxon_names"].append(
                    {"name": name, "class": name_class}
                )


def stream_nodes(nodes, roots):
    """Add lineage info and stream taxonomy nodes."""
    for taxon_id, obj in nodes.items():
        lineage = obj.copy()
        lineage.update(
            {
                "lineage": [
                    {
                        "taxon_id": obj["taxon_id"],
                        "taxon_rank": obj["taxon_rank"],
                        "scientific_name": obj["scientific_name"],
                        "node_depth": 0,
                    }
                ]
            }
        )
        descendant = False
        if taxon_id in roots:
            descendant = True
        depth = 0
        while "parent" in obj and obj["parent"] in nodes:
            depth += 1
            parent = obj["parent"]
            obj = nodes[parent]
            lineage["lineage"].append(
                {
                    "taxon_id": obj["taxon_id"],
                    "taxon_rank": obj["taxon_rank"],
                    "scientific_name": obj["scientific_name"],
                    "node_depth": depth,
                }
            )
            if obj["taxon_id"] in roots:
                descendant = True
                break
        if descendant:
            yield "taxon-%s" % taxon_id, lineage


def parse_ncbi_taxdump(path, root=None):
    """Expand lineages from nodes dict."""
    if root is None:
        root = ["1"]
    if not isinstance(root, list):
        root = [root]
    roots = list(map(str, root))
    nodes = parse_ncbi_nodes_dmp("%s/nodes.dmp" % path)
    parse_ncbi_names_dmp("%s/names.dmp" % path, nodes)
    yield from stream_nodes(nodes, roots)


def add_xrefs(names, xrefs):
    """Add xrefs to a list of taxon names."""
    dbs = {
        "gbif": {"source": "GBIF", "stub": "https://www.gbif.org/species/"},
        "irmng": {
            "source": "IRMNG",
            "stub": "https://www.irmng.org/aphia.php?p=taxdetails&id=",
        },
        "ott": {
            "source": "OTT",
            "stub": "https://tree.opentreeoflife.org/opentree/argus/ottol@",
        },
        "ncbi": {
            "source": "NCBI",
            "stub": "https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=",
        },
        "silva": {
            "source": "SILVA",
            "stub": "https://www.arb-silva.de/browser/ssu-138.1/",
        },
        "worms": {
            "source": "WoRMS",
            "stub": "http://www.marinespecies.org/aphia.php?p=taxdetails&id=",
        },
    }
    for xref in xrefs:
        source, accession = xref.split(":")
        if source in dbs:
            names.append(
                {
                    "name": accession,
                    "class": "xref",
                    "source": dbs[source]["source"],
                    "source_url_stub": dbs[source]["stub"],
                }
            )


def parse_ott_nodes_dmp(path):
    """Parse Open tree of Life taxonomy.tsv file."""
    nodes = {}
    with open_file_handle(path) as fh:
        for line in fh:
            taxon_id, parent, taxon_name, taxon_rank, source_info, *_ignore = re.split(
                r"\s*\|\s*", line
            )
            if taxon_id == "uid":
                continue
            xrefs = ["ott:%s" % taxon_id]
            taxon_id = "ott_%s" % taxon_id
            nodes[taxon_id] = {
                "parent": "ott_%s" % parent,
                "taxon_rank": taxon_rank,
                "taxon_names": [],
            }
            if taxon_name:
                nodes[taxon_id].update(
                    {
                        "taxon_id": taxon_id,
                        "scientific_name": taxon_name,
                    }
                )
                nodes[taxon_id]["taxon_names"].append(
                    {"name": taxon_name, "class": "scientific_name"}
                )
            if source_info:
                xrefs += [re.sub(r"#\d+", "", xref) for xref in source_info.split(",")]
            add_xrefs(nodes[taxon_id]["taxon_names"], xrefs)
    return nodes


def parse_ott_names_dmp(path, nodes):
    """Parse synonyms.tsv file and add to nodes dict."""
    with open_file_handle(path) as fh:
        for line in fh:
            name, taxon_id, name_class, *_ignore = re.split(r"\s*\|\s*", line)
            if taxon_id == "uid":
                continue
            taxon_id = "ott_%s" % taxon_id
            if taxon_id in nodes:
                nodes[taxon_id]["taxon_names"].append(
                    {"name": name, "class": name_class}
                )


def parse_ott_taxdump(path, root=None):
    """Expand lineages from nodes dict."""
    if root is None:
        root = ["ott_805080"]
    if not isinstance(root, list):
        root = [root]
    roots = list(map(str, root))
    nodes = parse_ott_nodes_dmp("%s/taxonomy.tsv" % path)
    parse_ott_names_dmp("%s/synonyms.tsv" % path, nodes)
    yield from stream_nodes(nodes, roots)


def parse_taxonomy(taxonomy_type, path, root=None):
    """Parse taxonomy into list of dicts."""
    parsers = {"ncbi": parse_ncbi_taxdump, "ott": parse_ott_taxdump}
    parser = parsers.get(taxonomy_type.lower(), None)
    if parser is None:
        return None
    return parser(path, root)



#!/usr/bin/env python3
"""
GenomeHubs.

Usage: genomehubs [<command>] [<args>...] [-h|--help] [--version]

Commands:
    index           index a file
    init            initialise a new genomehub
    launch          launch a web service
    parse           parse a local or remote data source
    remove          remove one or more indices from a genomehub
    run             run an analysis
    search          search indexed files
    -h, --help      show this
    -v, --version   show version number
See 'genomehubs <command> --help' for more information on a specific command.
"""

from __future__ import annotations

import sys
import warnings
from datetime import datetime

from docopt import DocoptExit
from docopt import docopt

from genomehubs.vendor.tolkein import tolog

try:
    from importlib.metadata import entry_points
except ImportError:
    # Python < 3.8 fallback
    from importlib_metadata import entry_points

from .lib.version import __version__

LOGGER = tolog.logger(__name__)


# Python EOL dates from https://devguide.python.org/versions/
PYTHON_EOL_DATES = {
    (3, 9): datetime(2025, 10, 5),
    (3, 10): datetime(2026, 10, 4),
    (3, 11): datetime(2027, 10, 24),
    (3, 12): datetime(2028, 10, 2),
    (3, 13): datetime(2029, 10, 1),
}

# Minimum supported Python version
MIN_PYTHON_VERSION = (3, 10)


def check_python_version():
    """Check for deprecated or EOL Python versions and warn users."""
    current_version = (sys.version_info.major, sys.version_info.minor)
    current_date = datetime.now()

    # Check if running on unsupported version
    if current_version < MIN_PYTHON_VERSION:
        warnings.warn(
            f"Python {current_version[0]}.{current_version[1]} is no longer supported by GenomeHubs. "
            f"Please upgrade to Python {MIN_PYTHON_VERSION[0]}.{MIN_PYTHON_VERSION[1]} or later.",
            DeprecationWarning,
            stacklevel=3,
        )
        return

    # Check if running on version approaching EOL
    if current_version in PYTHON_EOL_DATES:
        eol_date = PYTHON_EOL_DATES[current_version]
        days_until_eol = (eol_date - current_date).days

        if days_until_eol < 0:
            # Already past EOL
            warnings.warn(
                f"Python {current_version[0]}.{current_version[1]} reached end-of-life on {eol_date.strftime('%Y-%m-%d')}. "
                "GenomeHubs support will be removed in a future release. "
                f"Please upgrade to Python {MIN_PYTHON_VERSION[0]}.{MIN_PYTHON_VERSION[1]} or later for continued support and security updates.",
                DeprecationWarning,
                stacklevel=3,
            )
        elif days_until_eol < 180:  # Warn 6 months before EOL
            warnings.warn(
                f"Python {current_version[0]}.{current_version[1]} will reach end-of-life on {eol_date.strftime('%Y-%m-%d')} "
                f"({days_until_eol} days). "
                f"Please plan to upgrade to Python {MIN_PYTHON_VERSION[0]}.{MIN_PYTHON_VERSION[1]} or later.",
                FutureWarning,
                stacklevel=3,
            )


def cli():
    """Entry point."""
    check_python_version()
    if len(sys.argv) > 1:
        try:
            args = docopt(__doc__, help=False, version=__version__)
        except DocoptExit:
            args = {"<command>": sys.argv[1]}
        if args["<command>"]:
            # load <command> from entry_points
            eps = entry_points()
            # Handle both old and new entry_points() API
            if hasattr(eps, "select"):
                # New API (Python 3.10+)
                subcommand_eps = eps.select(group="genomehubs.subcmd")
            else:
                # Old API (Python 3.8-3.9)
                subcommand_eps = eps.get("genomehubs.subcmd", [])

            for entry_point in subcommand_eps:
                if entry_point.name == args["<command>"]:
                    subcommand = entry_point.load()
                    sys.exit(subcommand())
            LOGGER.error("'genomehubs %s' is not a valid command", args["<command>"])
            sys.exit(1)
    print(__doc__)
    raise DocoptExit



#!/usr/bin/env python3

"""
Search for entries.

Usage:
    genomehubs search [--assembly STRING...] [--feature STRING...]
                [--meta STRING...] [--overlap STRING...]
                [--return STRING...] [--taxon STRING...]
                [--gene-tree-node STRING...]
                [--species-tree-node STRING...]
                [--configfile YAML...] [--taxonomy-root INT]
                [--es-host HOSTNAME] [--es-port PORT]

Options:
    --assembly STRING           Assembly name or accession.
    --feature STRING            Feature value to search.
    --meta STRING               Metadata value to search.
    --overlap STRING            Feature to overlap.
    --return STRING             Formats to return
    --taxon STRING              Taxon or clade name or taxid.
    --gene-tree-node STRING     Gene tree node name or ID.
    --species-tree-node STRING  Tree name or ID.
    --configfile YAML           YAML configuration file.
    --taxonomy-root INT         Root taxid for taxonomy index.
    --es-host HOSTNAME          Elasticseach hostname/URL.
    --es-port PORT              Elasticseach port number.
    -h, --help                  Show this
    -v, --version               Show version number

Examples:
    # 1. All data for a clade
        ./genomehubs search \
            --taxon Nymphalidae \
            --target RAW

    # 2. All butterfly genomes above contig N50 1M
        ./genomehubs search \
            --taxon Lepidoptera \
            --meta 'assembly.contig_n50>1M' \
            --return FASTA,GFF3

    # 3. All genes with a given interpro domain
        ./genomehubs search \
            --assembly Hmel2.5 \
            --feature interpro_accession=IPR123456 \
            --overlap gene \
            --return GFF3

    # 4. All genes in a clade which are single copy busco genes"
        ./genomehubs search \
            --taxon Lepidoptera \
            --feature busco_status=Complete|Fragmented \
            --overlap gene \
            --return protein.FASTA
"""


import sys

from docopt import docopt

from .version import __version__


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index("search") + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    print(args)


if __name__ == "__main__":
    cli()



"""Evolutionary informatics package."""

from .lib import tobin
from .lib import tofetch
from .lib import tofile
from .lib import toinsdc
from .lib import tolog
from .lib import totax
from .lib import version
from .tolkein import cli



#!/usr/bin/env python3
"""INSDC methods."""

import re
from collections import defaultdict
from urllib.parse import urlencode

import ujson
import xmltodict
from tqdm import tqdm

from .tofetch import fetch_stream
from .tofetch import fetch_url

PORTAL = "https://www.ebi.ac.uk/ena/portal/api"


def count_taxon_assembly_meta(root):
    """
    Count INSDC assemblies descended from root taxon.

    Args:
        root (int): Root taxon taxid.

    Returns:
        int: Count of assemblies for taxa descended from root. Will return None on error.
    """
    url = '%s/count?query="tax_tree(%s)"&result=assembly&format=json' % (
        PORTAL,
        str(root),
    )
    count = fetch_url(url)
    if count is None:
        return None
    return int(count)


def fetch_wgs_assembly_meta(root, *, count=-1, offset=0, page=10000):
    """
    Query INSDC WGS assemblies descended from root taxon.

    Args:
        root (int): Root taxon taxid.
        count (int): Number of assemblies to return.
            Default value (-1) returns all assemblies.
        offset (int): Offset of first assembly to return. Defaults to 0.
        page (int): Number of assemblies to fetch per API request. Defaults to 10000.

    Yields:
        dict: A dict of INSDC WGS assembly metadata keyed on sample accession.
    """
    fields = {
        "accession": "wgs_accession",
        "host": "host_scientific_name",
        "first_public": "wgs_first_public",
        "last_updated": "wgs_last_updated",
        "location": "sample_location",
        "study_accession": "study_accession",
        "sample_accession": "sample_accession",
        "sex": "sample_sex",
    }
    options = {
        "fields": ",".join(fields.keys()),
        "format": "json",
        "offset": offset,
        "limit": page,
        "query": '"tax_tree(%d)"' % int(root),
        "result": "wgs_set",
    }
    returned = 0
    wgs_meta = {}
    while returned < count or count == -1:
        url = "%s/search?%s" % (PORTAL, urlencode(options))
        batch_meta = fetch_url(url)
        if not batch_meta or batch_meta is None:
            if not wgs_meta:
                wgs_meta = None
            break
        for entry in ujson.loads(batch_meta):
            wgs_meta[entry["sample_accession"]] = {
                fields[key]: value for key, value in entry.items()
            }
        returned += page
        options["offset"] += page
    return wgs_meta


def stream_taxon_assembly_meta(root, *, count=-1, offset=0, page=10000):
    """
    Query INSDC assemblies descended from root taxon.

    Args:
        root (int): Root taxon taxid.
        count (int): Number of assemblies to return.
            Default value (-1) returns all assemblies.
        offset (int): Offset of first assembly to return. Defaults to 0.
        page (int): Number of assemblies to fetch per API request. Defaults to 10000.

    Yields:
        dict: Normalised dict of INSDC metadata.
    """
    fields = {
        "accession": "gca_accession",
        "study_accession": "study_accession",
        "sample_accession": "sample_accession",
        "secondary_sample_accession": "secondary_sample_accession",
        "assembly_name": "assembly_name",
        "assembly_title": "assembly_title",
        "study_name": "study_name",
        "study_title": "study_title",
        "study_description": "study_description",
        "tax_id": "taxon_id",
        "scientific_name": "scientific_name",
        "strain": "assembled_strain",
        "base_count": "assembly_span",
        "assembly_level": "assembly_level",
        "genome_representation": "genome_representation",
        "last_updated": "last_updated",
        "version": "assembly_version",
        "assembly_type": "assembly_type",
    }
    options = {
        "fields": ",".join(fields.keys()),
        "format": "json",
        "offset": offset,
        "limit": page,
        "query": '"tax_tree(%d)"' % int(root),
        "result": "assembly",
    }
    wgs_meta = fetch_wgs_assembly_meta(root)
    returned = 0
    while returned < count or count == -1:
        url = "%s/search?%s" % (PORTAL, urlencode(options))
        batch_meta = fetch_url(url)
        if not batch_meta or batch_meta is None:
            break
        for entry in ujson.loads(batch_meta):
            entry_meta = {fields[key]: value for key, value in entry.items()}
            if entry_meta["sample_accession"] in wgs_meta:
                entry_meta = {**entry_meta, **wgs_meta[entry_meta["sample_accession"]]}
            yield entry_meta
        returned += page
        options["offset"] += page



#!/usr/bin/env python3

"""Assembly and Sample indexing methods."""

import copy
from collections import defaultdict

from genomehubs.vendor.tolkein import tolog

from .es_functions import document_by_id
from .es_functions import query_keyword_value_template
from .hub import add_attribute_values
from .hub import chunks
from .hub import index_templator
from .taxon import add_taxonomy_info_to_meta
from .taxon import find_or_create_taxa

LOGGER = tolog.logger(__name__)


def index_template(taxonomy_name, opts, *, index_type="sample"):
    """Index template (includes name, mapping and types)."""
    parts = [index_type, taxonomy_name, opts["hub-name"], opts["hub-version"]]
    template = index_templator(parts, opts)
    return template


def get_list_entries_by_dict_value(values, list_of_dicts, *, key="key"):
    """Get entries from a list of dicts by key."""
    entries = []
    for entry in list_of_dicts:
        if key in entry and entry[key] in values:
            entries.append(entry)
    return entries


def get_list_indices_by_dict_value(values, list_of_dicts, *, key="key"):
    """Get indices from a list of dicts by key."""
    indices = []
    for idx, entry in enumerate(list_of_dicts):
        if key in entry and entry[key] in values:
            indices.append(idx)
    return indices


def add_identifiers_to_list(existing, new, *, blanks=set({"NA"})):
    """Add identifiers to a list if they do not already exist."""
    # TODO: include source?
    identifiers = defaultdict(dict)
    for entry in existing:
        identifiers[entry["class"]][entry["identifier"]] = True
    for entry in new:
        id_class = entry["class"]
        identifier = entry["identifier"]
        if (
            identifier not in blanks
            and id_class not in identifiers
            and identifier not in identifiers[id_class]
        ):
            existing.append(
                {
                    "identifier": identifier,
                    "class": id_class,
                    **{
                        key: value
                        for key, value in entry.items()
                        if key.startswith("source_") and not key.endswith("_template")
                    },
                }
            )
            identifiers[id_class][identifier] = True


def lookup_docs_by_doc_id(
    es, values, template, *, return_type="list", index_type="sample"
):
    """Retrieve existing taxa from index."""
    docs = []
    if return_type == "dict":
        docs = {}
    res = document_by_id(es, values, template["index_name"])
    for key, value in res.items():
        key = key.replace(f"{index_type}-", "")
        if return_type == "list":
            docs.append(key)
        else:
            docs.update({key: value})
    return docs


def add_identifiers_and_attributes_to_entries(
    es, data, opts, *, template, taxon_template, blanks=set(["NA"]), index_type="sample"
):
    """Add identifiers and attributes to entries."""
    all_entries = {}
    taxon_id_by_entry = {}
    for taxon_id, entries in data.items():
        for entry in entries:
            entry["taxon_id"] = taxon_id
            for identifier in entry["identifiers"]:
                if identifier["class"] == f"{index_type}_id":
                    all_entries[identifier["identifier"]] = entry
                    taxon_id_by_entry[identifier["identifier"]] = taxon_id
    for values in chunks(list(all_entries.keys()), 500):
        res = query_keyword_value_template(
            es,
            "attributes_identifiers_by_keyword_value",
            f"{index_type}_id",
            values,
            index=template["index_name"],
        )
        if res is None:
            res = {"responses": [{} for value in values]}
        taxon_ids = set()
        for value in values:
            taxon_ids.add(taxon_id_by_entry[value])
        taxa = find_or_create_taxa(
            es,
            opts,
            taxon_ids=taxon_ids,
            taxon_template=taxon_template,
            asm_by_taxon_id=data,
        )
        for index, response in enumerate(res["responses"]):
            doc = {
                "_id": f"{index_type}-{values[index]}",
                "_source": {
                    f"{index_type}_id": values[index],
                    "identifiers": [],
                    "attributes": [],
                },
            }
            if "hits" in response and response["hits"]["total"]["value"] == 1:
                doc = response["hits"]["hits"][0]
            doc_data = all_entries[doc["_source"][f"{index_type}_id"]]
            identifiers = copy.deepcopy(doc["_source"]["identifiers"])
            if "identifiers" in doc_data:
                identifiers = identifiers + doc_data["identifiers"]
            add_identifiers_to_list(
                doc["_source"]["identifiers"], identifiers, blanks=blanks
            )
            if "attributes" not in doc["_source"] or not doc["_source"]["attributes"]:
                doc["_source"]["attributes"] = []
            if "attributes" in doc_data:
                add_attribute_values(
                    doc["_source"]["attributes"], doc_data["attributes"], raw=False
                )
            doc["_source"]["taxon_id"] = doc_data["taxon_id"]
            try:
                add_taxonomy_info_to_meta(
                    doc["_source"], taxa[doc_data["taxon_id"]]["_source"]
                )
                yield doc["_id"], doc["_source"]
            except KeyError:
                LOGGER.warning(
                    "Taxon ID %s was not found in the taxonomy" % doc_data["taxon_id"]
                )


def add_entry_attributes_to_taxon(
    batch,
    entry_by_taxon_id,
    taxa,
    *,
    index_name,
    shared_attributes,
    index_type="sample",
):
    """Add entry attributes to taxon."""
    for taxon_id, entries in entry_by_taxon_id.items():
        if taxon_id in taxa:
            taxon_attributes = defaultdict(list)
            if "attributes" not in taxa[taxon_id]:
                taxa[taxon_id]["attributes"] = []
            for idx in entries:
                entry_meta = batch[idx]
                attributes = get_list_entries_by_dict_value(
                    shared_attributes, entry_meta["attributes"]
                )
                for attr in attributes:
                    taxon_attr = {**attr}
                    del taxon_attr["key"]
                    taxon_attr["source_index"] = index_name
                    taxon_attr["source_doc_id"] = (
                        f"{index_type}-{entry_meta[f'{index_type}_id']}"
                    )
                    taxon_attributes[attr["key"]].append(taxon_attr)
            for key, values in taxon_attributes.items():
                indices = get_list_indices_by_dict_value(
                    {key}, taxa[taxon_id]["attributes"]
                )
                if len(indices) == 1:
                    idx = indices[0]
                    # TODO: test if values are already present
                    taxa[taxon_id]["attributes"][idx]["values"] += values
                else:
                    taxa[taxon_id]["attributes"].append({"key": key, "values": values})


def collate_unique_key_value_indices(key, list_of_dicts):
    """Collate indices of unique key values in a list of dicts."""
    unique_key_values = set()
    entry_indices_by_key_value = defaultdict(list)
    for idx, entry in enumerate(list_of_dicts):
        if entry[key] not in unique_key_values:
            unique_key_values.add(entry[key])
        entry_indices_by_key_value[entry[key]].append(idx)
    return list(unique_key_values), entry_indices_by_key_value



#!/usr/bin/env python3

"""File indexing methods."""


import contextlib
import hashlib
import os
import urllib
from datetime import date
from mimetypes import guess_type
from pathlib import Path
from shutil import copyfile
from traceback import format_exc

import filetype
from PIL import Image

from genomehubs.vendor.tolkein import tofetch
from genomehubs.vendor.tolkein import tofile
from genomehubs.vendor.tolkein import tolog

from .analysis import index_template as analysis_index_template
from .es_functions import document_by_id
from .es_functions import index_stream
from .hub import index_templator
from .sample import index_template as entry_index_template
from .sample import lookup_docs_by_doc_id
from .taxon import find_or_create_taxa
from .taxon import index_template as taxon_index_template
from .taxon import list_ancestors

LOGGER = tolog.logger(__name__)

TODAY = date.today().strftime("%Y-%m-%d")


def index_template(taxonomy_name, opts):
    """Index template (includes name, mapping and types)."""
    parts = ["file", taxonomy_name, opts["hub-name"], opts["hub-version"]]
    return index_templator(parts, opts)


def force_copy(sourcename, destname, symlink=True):
    """Create or replace a symlink to a file."""
    localfile = Path(destname)
    os.makedirs(Path(destname).parent, exist_ok=True)
    with contextlib.suppress(FileNotFoundError):
        localfile.unlink()
    if symlink:
        localfile.symlink_to(Path(sourcename))
    else:
        copyfile(Path(sourcename), localfile)


def process_image_file(infile, filename, opts, *, dest_dir="./", attrs=None):
    """Process an image file for indexing."""
    if attrs is None:
        attrs = {}
    filepath = Path(filename)
    dimensions = (100, 100)
    thumbname = f"{filepath.stem}.thm{filepath.suffix}"
    with Image.open(infile) as im:
        attrs.update(
            {
                "preview_name": thumbname,
                "preview_mime_type": attrs["mime_type"],
                "format": im.format,
                "size_pixels": f"{im.width}x{im.height}",
            }
        )
        if filename != thumbname:
            try:
                im.thumbnail(dimensions)
                os.makedirs(f'{opts["hub-path"]}/{dest_dir}', exist_ok=True)
                im.save(f'{opts["hub-path"]}/{dest_dir}/{thumbname}')
            except OSError as err:
                print(err)
                LOGGER.warn("Cannot create thumbnail for '%s'", infile)
        attrs.update()
    return attrs


def set_file_meta_defaults(opts):
    """Set default values for file metadata."""
    defaults = {}
    for key in {
        "taxon-id",
        "assembly-id",
        "analysis-id",
        "file-title",
        "file-description",
    }:
        if value := opts.get(key, False):
            defaults[key.replace("file-", "").replace("-", "_")] = value
    return defaults


def make_local_copy(infile, filepath, filename, attrs, opts, local):
    """Link or copy file to local directory."""
    try:
        dest_dir = str(filepath.parent.relative_to(opts["hub-path"]))
    except ValueError:
        dest_dir = "files"
        try:
            dest_dir += "/taxon-" + attrs.get("taxon_id", "NA")
        except TypeError:
            dest_dir += "/taxon-MULTI"
        try:
            dest_dir += "/assembly-" + attrs.get("assembly_id", "NA")
        except TypeError:
            dest_dir += "/assembly-MULTI"
        dest_dir += "/analysis-" + attrs.get("analysis_id", "all")
        localname = f'{opts["hub-path"]}/{dest_dir}/{filename}'
        if local == "symlink":
            force_copy(infile, localname)
        elif local == "copy":
            force_copy(infile, localname, False)
    return dest_dir


def process_file(
    infile,
    opts,
    *,
    file_template=None,
    analysis_template=None,
    filename=None,
    meta=None,
    local="symlink",
):
    """Process a file for indexing."""
    filepath = Path(infile)
    if filename is None:
        filename = filepath.name
    if meta is None:
        meta = {}
    attrs = {
        "name": filename,
        "size_bytes": os.path.getsize(infile),
        **set_file_meta_defaults(opts),
        **meta,
    }
    dest_dir = make_local_copy(infile, filepath, filename, attrs, opts, local)
    attrs["location"] = dest_dir
    kind = filetype.guess(infile)
    if kind is not None:
        attrs.update({"extension": kind.extension, "mime_type": kind.mime})
        if filetype.is_image(infile):
            process_image_file(infile, filename, opts, dest_dir=dest_dir, attrs=attrs)
        # TODO: #32 Generate text file previews
    else:
        attrs["extension"] = filepath.suffix.replace(".", "")
        attrs["mime_type"] = guess_type(filepath)[0]
    file_props = file_template["mapping"]["mappings"]["properties"].keys()
    analysis_props = analysis_template["mapping"]["mappings"]["properties"].keys()
    file_attrs = {}
    analysis_attrs = {}
    # split attrs into 2 sets for indexing
    for key, value in attrs.items():
        if key in file_props:
            file_attrs[key] = value
            if key == "analysis_id":
                analysis_attrs[key] = value
        elif key in analysis_props:
            analysis_attrs[key] = value
        elif key == "analysis":
            for sub_key, sub_value in value.items():
                if sub_key in analysis_props:
                    analysis_attrs[sub_key] = sub_value
    file_location = f'{file_attrs["location"]}/{file_attrs["name"]}'
    file_attrs["file_id"] = hashlib.md5(file_location.encode("utf-8")).hexdigest()
    if not file_attrs.get("date_created", False):
        file_attrs["date_created"] = TODAY
    if "url" in file_attrs:
        file_attrs["date_accessed"] = TODAY
    return file_attrs, analysis_attrs


def index_files(es, files, taxonomy_name, opts):
    """Index files."""
    file_template = index_template(taxonomy_name, opts)
    analysis_template = analysis_index_template(taxonomy_name, opts)
    for infile in files:
        LOGGER.info("Indexing %s", infile)
        p = Path(infile)
        if p.is_dir():
            LOGGER.warn(
                "Argument to --file must be a valid file path. '%s' is a directory",
                infile,
            )
            continue
        file_attrs, analysis_attrs = process_file(
            infile,
            opts,
            file_template=file_template,
            analysis_template=analysis_template,
        )


def update_analysis_attributes(es, analyses, analysis_attrs, analysis_template):
    """Update analysis attributes ready for reindexing."""
    analysis_id = "analysis-%s" % analysis_attrs["analysis_id"]
    action = "index"
    if analysis_id not in analyses:
        res = document_by_id(es, analysis_id, analysis_template["index_name"])
        if res is not None and analysis_id in res:
            analyses.update({analysis_id: res[analysis_id]})
    if analysis_id in analyses:
        action = "update"
        taxon_id = analysis_attrs.get("taxon_id", None)
        assembly_id = analysis_attrs.get("assembly_id", None)
        if (
            taxon_id
            and taxon_id is not None
            and analyses[analysis_id].get("taxon_id", taxon_id) != taxon_id
        ):
            action = "mismatch"
        elif (
            assembly_id
            and assembly_id is not None
            and analyses[analysis_id].get("assembly_id", assembly_id) != assembly_id
        ):
            action = "mismatch"
        if action == "mismatch":
            LOGGER.warning(
                "taxon/assembly ID mismatch for %s, unable to import associated files",
                analysis_id,
            )
            return False
        LOGGER.info("Updating %s", analysis_id)
        # update analysis
        for key, value in analysis_attrs.items():
            if value != "" and value != []:
                if key not in analyses[analysis_id]:
                    analyses[analysis_id].update({key: value})
                elif value != analyses[analysis_id][key]:
                    LOGGER.info(
                        "Updating %s value for %s to '%s'", key, analysis_id, value
                    )
    else:
        analyses[analysis_id] = {}
        for key, value in analysis_attrs.items():
            if value != "":
                analyses[analysis_id].update({key: value})
        # set file counter
        analyses[analysis_id].update({"file_count": 0})
    return action


def update_file_attributes(es, files, file_attrs, analyses, file_template):
    """Update file attributes ready for reindexing."""
    file_id = "file-%s" % file_attrs["file_id"]
    action = "index"
    if file_id not in files:
        res = document_by_id(
            es,
            file_id,
            file_template["index_name"],
        )
        if res is not None and file_id in res:
            files.update({file_id: res[file_id]})
    if file_id in files:
        action = "update"
        analysis_id = file_attrs.get("analysis_id", None)
        if files[file_id].get("analysis_id", analysis_id) != analysis_id:
            LOGGER.warning(
                "analysis ID mismatch for %s, unable to import %s",
                file_id,
                file_attrs.get("name", None),
            )
            return False
        LOGGER.info("Updating %s", file_id)
        # update analysis
        for key, value in file_attrs.items():
            if value != "":
                if key not in files[file_id]:
                    files[file_id].update({key: value})
                elif value != files[file_id][key]:
                    LOGGER.info("Updating %s value for %s to '%s'", key, file_id, value)
    else:
        files[file_id] = {}
        for key, value in file_attrs.items():
            if value != "":
                files[file_id].update({key: value})
        analyses["analysis-%s" % file_attrs["analysis_id"]]["file_count"] += 1
    return action


def stream_docs(doc_dict):
    """Stream a dict of docs."""
    for doc_id, doc in doc_dict.items():
        yield doc_id, doc


def index_docs(es, opts, doc_collection, template, *, dry_run=False):
    """Index a collection of docs."""
    for op_type, doc_dict in doc_collection.items():
        if doc_dict:
            docs = stream_docs(doc_dict)
            index_stream(
                es,
                template["index_name"],
                docs,
                _op_type=op_type,
                dry_run=dry_run,
                log=opts.get("log-es", True),
                chunk_size=opts.get("es-batch", 500),
            )


def check_taxa_exist(es, attrs, taxonomy_name, opts):
    """Check that the specified taxon_id(s) exist in the taxon index."""
    taxon_template = taxon_index_template(taxonomy_name, opts)
    taxon_ids = attrs.get("taxon_id", False)
    if not taxon_ids:
        return {}
    if not isinstance(taxon_ids, list):
        taxon_ids = [taxon_ids]
    taxa = {}
    if taxon_ids:
        taxa = find_or_create_taxa(
            es, opts, taxon_ids=taxon_ids, taxon_template=taxon_template
        )
    if len(taxon_ids) > len(taxa.keys()):
        found = taxa.keys()
        missing = ", ".join(
            [taxon_id for taxon_id in taxon_ids if taxon_id not in found]
        )
        LOGGER.warn("Taxon ID(s) '%s' not found in the specified taxonomy.", missing)
        LOGGER.warn("Unable to import analysis '%s'.", attrs["analysis_id"])
        return None
    return taxa


def check_entries_exist(es, attrs, taxonomy_name, opts, *, index_type="assembly"):
    """Check that the specified doc_id(s) exist in the {index_type} index."""
    # similar to check_taxa_exist but for assemblies
    entry_template = entry_index_template(taxonomy_name, opts, index_type=index_type)
    entry_ids = attrs.get(f"{index_type}_id", False)
    if not isinstance(entry_ids, list):
        entry_ids = [entry_ids]
    entries = {}
    if entry_ids:
        entries = lookup_docs_by_doc_id(
            es,
            [f"{index_type}-{entry_id}" for entry_id in entry_ids],
            entry_template,
            return_type="dict",
            index_type="assembly",
        )
    return entries


def index_metadata(es, file, taxonomy_name, opts, *, dry_run=False):
    """Index file metadata."""
    data = tofile.load_yaml(file)
    file_template = index_template(taxonomy_name, opts)
    analysis_template = analysis_index_template(taxonomy_name, opts)
    if data is None:
        LOGGER.warn(f"Unable to load file metadata from '{file}'")
        return
    analyses = {}
    files = {}
    analysis_docs = {"index": {}, "update": {}}
    file_docs = {"index": {}, "update": {}}
    for meta in data:
        try:
            local = "symlink"
            if "path" in meta:
                if meta["path"].startswith("~"):
                    infile = os.path.expanduser(meta["path"])
                else:
                    infile = meta["path"]
                del meta["path"]
            elif "url" in meta:
                try:
                    infile = tofetch.fetch_tmp_file(meta["url"])
                    local = None
                except urllib.error.HTTPError as err:
                    LOGGER.warn(f'Got {str(err.code)} error for {meta["url"]}')
                    continue
                except ConnectionResetError:
                    LOGGER.warn(f'Got ConnectionResetError for {meta["url"]}')
                    continue
            elif "name" in meta:
                infile = meta["name"]
            else:
                LOGGER.warn("Found a record with no associated file in '%s'" % file)
            if "name" in meta:
                filename = meta["name"]
                del meta["name"]
            else:
                filename = None
            local = meta.pop("local", local)
            file_attrs, analysis_attrs = process_file(
                infile,
                opts,
                file_template=file_template,
                analysis_template=analysis_template,
                filename=filename,
                meta=meta,
                local=local,
            )
            if "analysis_id" not in analysis_attrs:
                file_name = file_attrs.get("name", file_attrs["file_id"])
                LOGGER.warn(
                    "Unable to import %s: analysis_id must be specified", file_name
                )
                continue
            # TODO: #30 check taxon_id(s) and assembly_id(s) exist in database
            assemblies = check_entries_exist(
                es, analysis_attrs, taxonomy_name, opts, index_type="assembly"
            )
            if assemblies is None:
                continue
            if "taxon_id" in analysis_attrs and analysis_attrs["taxon_id"]:
                taxa = check_taxa_exist(es, analysis_attrs, taxonomy_name, opts)
            else:
                # set taxa from assembly index
                taxa = {
                    assembly["taxon_id"]: assembly for assembly in assemblies.values()
                }
                analysis_attrs["taxon_id"] = list(taxa.keys())
            if taxa is None:
                continue
            # TODO: #33 include lineage summary in analysis index
            analysis_attrs["ancestors"] = list_ancestors(taxa)
            # Fetch existing analysis entry if available
            action = update_analysis_attributes(
                es, analyses, analysis_attrs, analysis_template
            )
            if not action:
                # TODO: remove analysis and files with shared analysis name
                continue
            analysis_id = f'analysis-{analysis_attrs["analysis_id"]}'
            if analysis_id in analysis_docs["index"]:
                action = "index"
            analysis_docs[action].update({analysis_id: analyses[analysis_id]})
            # prepare files for indexing (include check for dupes and increment analysis file count)
            file_action = update_file_attributes(
                es, files, file_attrs, analyses, file_template
            )
            if not file_action:
                continue
            file_id = f'file-{file_attrs["file_id"]}'
            if analysis_id in analysis_docs["index"]:
                action = "index"
            file_docs[file_action].update({file_id: files[file_id]})
        except Exception:
            LOGGER.warn("Caught unexpected error")
            print(format_exc())
    # Create/update index entry according to returned action
    LOGGER.info("Indexing analyses")
    index_docs(es, opts, analysis_docs, analysis_template, dry_run=dry_run)
    LOGGER.info("Indexing files")
    index_docs(es, opts, file_docs, file_template, dry_run=dry_run)
    index_docs(es, opts, file_docs, file_template, dry_run=dry_run)



"""Stable public re-exports for selected tolkein helpers."""

from .vendor.tolkein import tofetch
from .vendor.tolkein import tofile
from .vendor.tolkein import tolog

__all__ = ["tofetch", "tofile", "tolog"]
